/*
   core/reputation.cc
   This file is part of the Osirion project and is distributed under
   the terms of the GNU General Public License version 2
*/

#include "core/reputation.h"
#include "core/application.h"

namespace core {
	
Reputation::Reputation()
{
	reputation_dirty = false;
}

Reputation::~Reputation()
{
	clear();
}

void Reputation::clear()
{
	for (FactionReps::iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		delete (*it);
		(*it) = 0;
	}
	reputation_factionreps.clear();
	
	reputation_dirty = false;
}

void Reputation::assign(const Reputation &other)
{
	clear();
	
	for (FactionReps::const_iterator it = other.reputation_factionreps.begin(); it != other.reputation_factionreps.end(); ++it) {
		
		FactionRep *factionrep = new FactionRep((*it)->faction(), (*it)->reputation());
		reputation_factionreps.push_back(factionrep);
		
	}

}

void Reputation::set_dirty(const bool dirty)
{
	reputation_dirty = dirty;
	if (reputation_dirty) {
		reputation_timestamp = game()->timestamp();
	}
}

void Reputation::set_timestamp(const unsigned long timestamp)
{
	reputation_timestamp = timestamp;
}

const Info *Reputation::find(const Info *faction) const
{
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		if ((*it)->faction() == faction) {
			return (*it)->faction();
		}
	}
	
	return 0;
}

const float Reputation::reputation(const Info *faction) const
{
	if (!faction) {
		return 0.0f;
	}
	
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		if ((*it)->faction() == faction) {
			return (*it)->reputation();
		}
	}
	
	return 0.0f;
}

void Reputation::set_faction(const std::string &label, const Info *faction)
{
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		if ((*it)->label().compare(label) == 0) {
			(*it)->set_faction(faction);
		}
	}
}

void Reputation::set_reputation(const std::string &label, const float reputation)
{
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		if (((*it)->faction() && (label.compare((*it)->faction()->label()) == 0)) || ((*it)->label().compare(label) == 0)) {
			if ((*it)->reputation() != reputation) {
				(*it)->set_reputation(reputation);
				reputation_dirty = true;
			}
			return;
		}
	}
	
	FactionRep *factionrep = new FactionRep(label, reputation);
	reputation_factionreps.push_back(factionrep);
	reputation_dirty = true;
}

void Reputation::set_reputation(const Info *faction, const float reputation)
{
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		if ((!(*it)->faction() && ((*it)->label().compare(faction->label()) == 0)) || ((*it)->faction() == faction)) {
			if ((*it)->reputation() != reputation) {
				(*it)->set_reputation(reputation);
				reputation_dirty = true;
			}
			return;
		}
	}
	
	FactionRep *factionrep = new FactionRep(faction, reputation);
	reputation_factionreps.push_back(factionrep);
	reputation_dirty = true;
}

void Reputation::receive_server_update(std::istream &is)
{
	unsigned long id;
	float reputation;

	while ((is >> id) && ( is >> reputation)) {
		// this is client side and has to create info records as required
		set_reputation(game()->request_info(id), reputation);
	}
}
	
void Reputation::serialize_server_update(std::ostream & os) const
{
	for (FactionReps::const_iterator it = reputation_factionreps.begin(); it != reputation_factionreps.end(); ++it) {
		os << (*it)->faction()->id() << " " << roundf((*it)->reputation()) << " ";
	}
}

} // namespace core